header
{
package mvcase.persistence.javaparser;
}

{
import java.util.*;
import mvcase.models.*;
}
/** Java 1.3 AST Recognizer Grammar
 *
 * Author: (see java.g preamble)
 *
 * This grammar is in the PUBLIC DOMAIN
 */
class JavaTreeParser extends TreeParser;
options {
   importVocab = Java;
}
{
   int tab = 0;
   int opcount = 0;
   Stack classesstack;
   Operation roperation;
   Parameter rparameter;
   Attribute rattribute;
   boolean operationsignature = false;
   Vector specifications;
   public JavaTreeParser(mvcase.models.Class p_rclass, Vector p_specifications)
   {
      classesstack = new Stack();
      p_rclass.clearAttributes();
      p_rclass.clearOperations();
      p_rclass.clearNestedClasses();
      classesstack.push(p_rclass);
      roperation = null;
      rattribute = null;
      rparameter = null;
      operationsignature = false;
      specifications = p_specifications;
   }
   public void incTab() { tab+=3; }
   public void decTab()
   {
      tab-=3;
      if(tab<0) tab=0;
   }
   public String getTab()
   {
      String ret = "";
      for(int i=0;i<tab;i++) ret+=" ";
      return ret;
   }

   public void startNestedClass()
   {
      mvcase.models.Class rclass = new mvcase.models.Class();
      classesstack.push(rclass);
   }

   public void endNestedClass()
   {
      mvcase.models.Class rclass = null;
      mvcase.models.Class containerclass = null;
      if(!classesstack.empty())
         rclass = (mvcase.models.Class)classesstack.pop();
      if(!classesstack.empty())
         containerclass = (mvcase.models.Class)classesstack.peek();
      if(rclass!=null && containerclass!=null)
         containerclass.addNestedClass(rclass);
   }

   public void setAsClass()
   {
      mvcase.models.Class rclass = null;
      if(!classesstack.empty())
         rclass = (mvcase.models.Class)classesstack.peek();
      if(rclass!=null)
      {
         if(rclass.isInterface())
            rclass.setStereotype("");
      }
   }

   public void setAsInterface()
   {
      mvcase.models.Class rclass = null;
      if(!classesstack.empty())
         rclass = (mvcase.models.Class)classesstack.peek();
      if(rclass!=null)
      {
         if(!rclass.isInterface())
            rclass.setStereotype("interface");
      }
   }

   public void setClassModifiers(String strmods)
   {
      mvcase.models.Class rclass = null;
      if(!classesstack.empty())
         rclass = (mvcase.models.Class)classesstack.peek();
      if(rclass!=null)
      {
         int index = strmods.indexOf("abstract");
         if(index!=-1)
            rclass.setAbstract(true);
      }
   }

   public void setClassDoc(String strdoc)
   {
      mvcase.models.Class rclass = null;
      if(!classesstack.empty())
         rclass = (mvcase.models.Class)classesstack.peek();
      if(rclass!=null)
      {
         rclass.setDoc(strdoc);
      }
   }

   public void setClassName(String strname)
   {
      mvcase.models.Class rclass = null;
      if(!classesstack.empty())
         rclass = (mvcase.models.Class)classesstack.peek();
      if(rclass!=null)
      {
         rclass.setName(strname);
      }
   }

   public void setClassExtends(String strextends)
   {
   }

   public void setClassImplements(String strimplements)
   {
   }

   public void addModifier(String mod)
   {
      if(roperation != null && operationsignature && rparameter == null)
      {
         addOperationModifier(mod);
      }
      else if(rattribute != null && roperation == null)
      {
         addAttributeModifier(mod);
      }
   }

   public void startOperationSignature(boolean reserved)
   {
      mvcase.models.Class rclass = null;
      if(!classesstack.empty())
         rclass = (mvcase.models.Class)classesstack.peek();
      if(rclass != null)
      {
         if(roperation == null)
         {
            roperation = new Operation();
            if(reserved)
               roperation.setStereotype("Reserved");
            operationsignature = true;
         }
      }
   }

   public void addOperationModifier(String mod)
   {
      if(mod.equals("synchronized")) roperation.setSynchronized(true);
      if(mod.equals("abstract")) roperation.setAbstract(true);
      if(mod.equals("native")) roperation.setNative(true);
      if(mod.equals("const")) roperation.setConst(true);
      if(mod.equals("static")) roperation.setStatic(true);
      if(mod.equals("final")) roperation.setFinal(true);
      if(mod.equals("public")) roperation.setExportControl("public");
      if(mod.equals("private")) roperation.setExportControl("private");
      if(mod.equals("protected")) roperation.setExportControl("protected");
   }

   public void addAttributeModifier(String mod)
   {
      if(mod.equals("static")) rattribute.setStatic(true);
      if(mod.equals("final")) rattribute.setFinal(true);
      if(mod.equals("public")) rattribute.setExportControl("public");
      if(mod.equals("private")) rattribute.setExportControl("private");
      if(mod.equals("protected")) rattribute.setExportControl("protected");
   }

   public void setOperationDoc(String doc)
   {
      if(roperation!=null && operationsignature)
      {
         if(doc!=null && !doc.trim().equals(""))
         {
            String sdoc = removeDocTabs(doc);
            if(sdoc.trim().startsWith("*** Operation ID="))
            {
               int index = sdoc.indexOf("\n");
               String squid = sdoc.trim();
               if(index!=-1)
               {
                  squid = sdoc.substring(0,index).trim();
                  sdoc = sdoc.substring(index+1);
               }
               else sdoc = null;
               int i1 = squid.indexOf("\"")+1;
               int i2 = squid.lastIndexOf("\"");
               squid = squid.substring(i1,i2);
               roperation.setQuid(squid);
            }
            roperation.setDoc(sdoc);
         }
      }
   }

   public void setOperationName(String name)
   {
      if(roperation!=null && operationsignature)
      {
         roperation.setName(name);
      }
   }

   public void setOperationType(String type)
   {
      if(roperation!=null && operationsignature)
      {
         roperation.setType(type);
      }
   }

   public void startParameter()
   {
      if(roperation!=null && operationsignature)
      {
         rparameter = new Parameter();
      }
   }

   public void endParameter(String name, String type)
   {
      if(roperation!=null && operationsignature && rparameter!=null)
      {
         rparameter.setName(name);
         rparameter.setType(type);
         roperation.addParameter(rparameter);
         rparameter = null;
      }
   }

   public void setOperationExceptions(String exc)
   {
      if(roperation!=null && operationsignature)
      {
         roperation.setExceptions(exc);
      }
   }

   public void endOperationSignature()
   {
      if(roperation != null)
         operationsignature = false;
   }

   public String getNextSpecification()
   {
      return (String)specifications.elementAt(opcount++);
   }

   public void setOperationBody(String body)
   {
      if(roperation != null && !operationsignature)
         roperation.setSpecification(removeTabs(body));
   }

   private String removeDocTabs(String str)
   {
      String ret = new String();
      StringTokenizer st = new StringTokenizer(str,"\r\n");
      while(st.hasMoreTokens())
      {
         String line = st.nextToken();
         ret += line.trim() + "\r\n";
      }
      return ret.trim();
   }

   private String removeTabs(String str)
   {
      String ret = new String();
      StringTokenizer st = new StringTokenizer(str,"\r\n");
      while(st.hasMoreTokens())
      {
         String line = st.nextToken();
         if(line.startsWith("      "))
            ret += line.substring(6) + "\r\n";
         else
            ret += line + "\r\n";
      }
      return ret;
   }

   public void addOperation()
   {
      mvcase.models.Class rclass = null;
      if(!classesstack.empty())
         rclass = (mvcase.models.Class)classesstack.peek();
      if(rclass!=null)
      {
         if(roperation!=null)
         {
            rclass.addOperation(roperation);
            roperation = null;
         }
      }
   }

   public void startAttribute(boolean reserved)
   {
      mvcase.models.Class rclass = null;
      if(!classesstack.empty())
         rclass = (mvcase.models.Class)classesstack.peek();
      if(rclass!=null && roperation == null)
      {
         if(rattribute==null)
         {
            rattribute = new Attribute();
            if(reserved)
               rattribute.setStereotype("Reserved");
         }
      }
   }

   public void setAttributeDoc(String doc)
   {
      if(rattribute!=null && roperation == null)
      {
         rattribute.setDoc(removeDocTabs(doc));
      }
   }

   public void endAttribute(String type, String name, String initial)
   {
      mvcase.models.Class rclass = null;
      if(!classesstack.empty())
         rclass = (mvcase.models.Class)classesstack.peek();
      if(rattribute!=null && rclass!=null && roperation == null)
      {
         rattribute.setType(type);
         rattribute.setName(name);
         if(initial!=null && initial.startsWith(" ="))
            rattribute.setInitialValue(initial.substring(2));
         else
            rattribute.setInitialValue(initial);
         rclass.addAttribute(rattribute);
         rattribute = null;
      }
   }
}

compilationUnit
	:
      (packageDefinition)?
      (importDefinition)*
      (typeDefinition)*
	;

packageDefinition 
   :  #( PACKAGE_DEF identifier )
	;

importDefinition
   :  #( IMPORT identifierStar )
	;

javadoccomments returns [ String doc ] { doc = ""; }
   :  (#( tk:JAVADOC_COMMENT { doc = tk.getText(); }))?
   ;
typeDefinition returns [ String strtypeDefinition ] { strtypeDefinition = ""; }
   :  #(CLASS_DEF { setAsClass(); }
        { String doc,aux1,aux2,aux3,aux4; }
        doc = javadoccomments { setClassDoc(doc); }
        aux1 = modifiers { setClassModifiers(aux1); }
        tkident1:IDENT { setClassName(tkident1.getText()); }
        aux2 = extendsClause { setClassExtends(aux2); }
        aux3 = implementsClause { setClassImplements(aux3); }
        aux4 = objBlock
        { strtypeDefinition = aux1+"class "+tkident1.getText()+" "+aux2+" "+aux3+" "+aux4; }
       )
   |  #(INTERFACE_DEF { setAsInterface(); }
        { String doc,aux1,aux2,aux3; }
        doc = javadoccomments { setClassDoc(doc); }
        aux1 = modifiers { setClassModifiers(aux1); }
        tkident2:IDENT { setClassName(tkident2.getText()); }
        aux2 = extendsClause { setClassExtends(aux2); }
        aux3 = interfaceBlock { setClassImplements(aux3); }
        { strtypeDefinition = aux1+"interface "+tkident2.getText()+" "+aux2+" "+aux3; }
       )
	;

typeSpec returns [ String strtypeSpec ] { strtypeSpec = ""; }
   :  #(TYPE strtypeSpec = typeSpecArray)
	;

typeSpecArray returns [ String strtypeSpecArray ] { strtypeSpecArray = ""; }
   :  #( ARRAY_DECLARATOR strtypeSpecArray = typeSpecArray { strtypeSpecArray += "[]"; })
   |  strtypeSpecArray = type
	;

type returns [ String strtype ] { strtype = ""; }
   : strtype = identifier
   | strtype = builtInType
	;

builtInType returns [ String strbuiltInType ] { strbuiltInType = ""; }
    :   "void" { strbuiltInType = "void"; }
    |   "boolean" { strbuiltInType = "boolean"; }
    |   "byte" { strbuiltInType = "byte"; }
    |   "char" { strbuiltInType = "char"; }
    |   "short" { strbuiltInType = "short"; }
    |   "int" { strbuiltInType = "int"; }
    |   "float" { strbuiltInType = "float"; }
    |   "long" { strbuiltInType = "long"; }
    |   "double" { strbuiltInType = "double"; }
    ;

modifiers returns [ String strmodifiers ] { strmodifiers = ""; }
   :  #( MODIFIERS ({ String aux; } aux=modifier { addModifier(aux); }{ strmodifiers+=aux+" ";})* )
	;

modifier returns [ String strmodifier ] { strmodifier = ""; }
    :   "private" { strmodifier = "private"; }
    |   "public" { strmodifier = "public"; }
    |   "protected" { strmodifier = "protected"; }
    |   "static" { strmodifier = "static"; }
    |   "transient" { strmodifier = "transient"; }
    |   "final" { strmodifier = "final"; }
    |   "abstract" { strmodifier = "abstract"; }
    |   "native" { strmodifier = "native"; }
    |   "threadsafe" { strmodifier = "threadsafe"; }
    |   "synchronized" { strmodifier = "synchronized"; }
    |   "const" { strmodifier = "const"; }
    |   "volatile" { strmodifier = "volatile"; }
    |   "strictfp" { strmodifier = "strictfp"; }
    ;

extendsClause returns [ String strextendsClause ] { strextendsClause = ""; }
   :  #(EXTENDS_CLAUSE ({ String aux; } aux=identifier { strextendsClause = "extends "+aux; })? )
	;

implementsClause returns [ String strimplementsClause ]
      { strimplementsClause = "";
        Vector impls = new Vector(); }
   :  #(IMPLEMENTS_CLAUSE
         ( { String aux; }
           aux=identifier
           {impls.add(aux); }
         )*
         { if(!impls.isEmpty())
           {
              strimplementsClause = "implements ";
              Iterator i = impls.iterator();
              while(i.hasNext())
              {
                 String str = (String)i.next();
                 strimplementsClause += str;
                 if(i.hasNext())
                    strimplementsClause += ",";
              }
           }
         }
       )
	;

interfaceBlock returns [ String strinterfaceBlock ] { strinterfaceBlock = ""; }
	:	#(	OBJBLOCK
         { strinterfaceBlock = "\r\n"+getTab()+"{\r\n"; }
         { incTab(); }
         { String aux = ""; }
         (  aux = methodDecl { strinterfaceBlock += getTab()+aux+";\r\n"; }
         |  aux = reservedMethodDecl { strinterfaceBlock += getTab()+aux+";\r\n"; }
         |  aux = variableDef { strinterfaceBlock += getTab()+aux+";\r\n"; }
         |  aux = reservedVariableDef { strinterfaceBlock += getTab()+aux+";\r\n"; }
			)*
         { decTab(); }
         { strinterfaceBlock += getTab()+"}\r\n"; }
		)
	;
	
objBlock returns [ String strobjBlock ] { strobjBlock = ""; }
	:	#(	OBJBLOCK
         { strobjBlock = "\r\n"+getTab()+"{\r\n"; }
         { incTab(); }
         { String aux = ""; }
         (  aux = ctorDef { strobjBlock += getTab()+aux; }
         |  aux = methodDef { strobjBlock += getTab()+aux; }
         |  aux = reservedMethodDef { strobjBlock += getTab()+aux; }
         |  aux = variableDef { strobjBlock += getTab()+aux+";\r\n"; }
         |  aux = reservedVariableDef { strobjBlock += getTab()+aux+";\r\n"; }
         |  { startNestedClass(); }aux = typeDefinition { endNestedClass(); } { strobjBlock += getTab()+aux; }
         |  #(STATIC_INIT
               { incTab(); }
               aux = slist
               { decTab(); }
               { strobjBlock += getTab()+aux; }
             )
         |  #(INSTANCE_INIT
               { incTab(); }
               aux = slist
               { decTab(); }
               { strobjBlock += getTab()+aux; }
             )
			)*
         { decTab(); }
         { strobjBlock += getTab()+"}"; }
		)
	;

ctorDef returns [ String strctorDef ] { strctorDef = ""; }
   :  #( CTOR_DEF
         { String doc,aux1,aux2,aux3; }
         { startOperationSignature(false); }
         { setOperationType(""); }
         doc = javadoccomments { setOperationDoc(doc); }
         aux1 = modifiers
         aux2 = methodHead
         { endOperationSignature(); }
         { incTab(); }
         ctorSList
         { aux3 = getNextSpecification(); }
         { decTab(); }
         { strctorDef = aux1+aux2; }
         { if(aux3!=null)
           {
              strctorDef += "\r\n"+getTab()+"{\r\n";
              { incTab(); }
              strctorDef += getTab()+aux3+"\r\n";
              { decTab(); }
              strctorDef += getTab()+"}\r\n";
           }
         }
         { setOperationBody(aux3); }
         { addOperation(); }
       )
	;

methodDecl returns [ String strmethodDecl ] { strmethodDecl = ""; }
   :  #( METHOD_DEF
         { String doc,aux1,aux2,aux3; }
         { startOperationSignature(false); }
         doc = javadoccomments { setOperationDoc(doc); }
         aux1 = modifiers
         aux2 = typeSpec
         { setOperationType(aux2); }
         aux3 = methodHead
         { endOperationSignature(); }
         { strmethodDecl = aux1+aux2+" "+aux3; }
         { addOperation(); }
       )
	;

methodDef returns [ String strmethodDef ] { strmethodDef = "strmethodDef"; }
   :  #( METHOD_DEF
         { String doc,aux1,aux2,aux3,aux4=null; }
         { startOperationSignature(false); }
         doc = javadoccomments { setOperationDoc(doc); }
         aux1 = modifiers
         aux2 = typeSpec
         { setOperationType(aux2); }
         aux3 = methodHead
         { endOperationSignature(); }
         { incTab(); }
         ( slist { aux4 = getNextSpecification(); })?
         { decTab(); }
         { strmethodDef = aux1+aux2+" "+aux3; }
         { if(aux4!=null)
           {
              strmethodDef += "\r\n"+getTab()+"{\r\n";
              incTab();
              strmethodDef += getTab()+aux4+"\r\n";
              decTab();
              strmethodDef += getTab()+"}\r\n";
           }
         }
         { setOperationBody(aux4); }
         { addOperation(); }
       )
	;

reservedMethodDecl returns [ String strmethodDecl ] { strmethodDecl = ""; }
   :  #( RESERVED_METHOD_DEF
         { String doc,aux1,aux2,aux3; }
         { startOperationSignature(true); }
         doc = javadoccomments { setOperationDoc(doc); }
         aux1 = modifiers
         aux2 = typeSpec
         { setOperationType(aux2); }
         aux3 = methodHead
         { endOperationSignature(); }
         { strmethodDecl = aux1+aux2+" "+aux3; }
         { addOperation(); }
       )
	;

reservedMethodDef returns [ String strmethodDef ] { strmethodDef = "strmethodDef"; }
   :  #( RESERVED_METHOD_DEF
         { String doc,aux1,aux2,aux3,aux4=null; }
         { startOperationSignature(true); }
         doc = javadoccomments { setOperationDoc(doc); }
         aux1 = modifiers
         aux2 = typeSpec
         { setOperationType(aux2); }
         aux3 = methodHead
         { endOperationSignature(); }
         { incTab(); }
         ( slist { aux4 = getNextSpecification(); })?
         { decTab(); }
         { strmethodDef = aux1+aux2+" "+aux3; }
         { if(aux4!=null)
           {
              strmethodDef += "\r\n"+getTab()+"{\r\n";
              incTab();
              strmethodDef += getTab()+aux4+"\r\n";
              decTab();
              strmethodDef += getTab()+"}\r\n";
           }
         }
         { setOperationBody(aux4); }
         { addOperation(); }
       )
	;

variableDef  returns [ String strvariableDef ] { strvariableDef = ""; }
   :  #( VARIABLE_DEF
         { String doc,aux1,aux2,aux3,aux4; }
         { startAttribute(false); }
         doc = javadoccomments { setAttributeDoc(doc); }
         aux1 = modifiers
         aux2 = typeSpec
         aux3 = variableDeclarator
         aux4 = varInitializer
         { endAttribute(aux2,aux3,aux4); }
         { strvariableDef = aux1+aux2+" "+aux3+aux4; }
       )
	;

reservedVariableDef  returns [ String strvariableDef ] { strvariableDef = ""; }
   :  #( RESERVED_VARIABLE_DEF
         { String doc,aux1,aux2,aux3,aux4; }
         { startAttribute(true); }
         doc = javadoccomments { setAttributeDoc(doc); }
         aux1 = modifiers
         aux2 = typeSpec
         aux3 = variableDeclarator
         aux4 = varInitializer
         { endAttribute(aux2,aux3,aux4); }
         { strvariableDef = aux1+aux2+" "+aux3+aux4; }
       )
	;

parameterDef returns [ String strparameterDef ] { strparameterDef = ""; }
   :  #( PARAMETER_DEF
          { String aux1,aux2; }
          { startParameter(); }
          aux1 = modifiers
          aux2 = typeSpec
          tkident:IDENT
          { endParameter(tkident.getText(),aux2); }
          { strparameterDef = aux1+aux2+" "+tkident.getText(); }
       )
	;

objectinitializer
   :  #(INSTANCE_INIT slist)
	;

variableDeclarator returns [ String strvariableDeclarator ] { strvariableDeclarator = ""; }
   :  tkident:IDENT { strvariableDeclarator = tkident.getText(); }
	|	LBRACK
      { String aux; } aux=variableDeclarator
      { strvariableDeclarator = aux+"[]"; }
	;

varInitializer returns [ String strvarInitializer ] { strvarInitializer = ""; }
   :  #(ASSIGN strvarInitializer = initializer) { strvarInitializer = " = "+strvarInitializer; }
	|
	;

initializer returns [ String strinitializer ] { strinitializer = ""; }
   :  strinitializer = expression
   |  strinitializer = arrayInitializer
	;

arrayInitializer returns [ String strarrayInitializer ] { strarrayInitializer = ""; }
   :  #( ARRAY_INIT
         { strarrayInitializer = "{ "; }
         ( { String aux; } aux = initializer { strarrayInitializer += aux + ","; } )*
         { strarrayInitializer = strarrayInitializer.substring(0,strarrayInitializer.length()-1); }
         { strarrayInitializer += " }"; }
       )
	;

methodHead returns [ String strmethodHead ] { strmethodHead = ""; }
   :  tkident:IDENT { strmethodHead = tkident.getText()+"("; }
      { setOperationName(tkident.getText()); }
      #( PARAMETERS
          { Vector pars = new Vector(); }
          ({ String aux1; } aux1 = parameterDef { pars.add(aux1); })*
          { if(!pars.isEmpty())
            {
               Iterator i = pars.iterator();
               while(i.hasNext())
               {
                  String par = (String)i.next();
                  strmethodHead += par;
                  if(i.hasNext())
                     strmethodHead += ",";
               }
            }
          }
       )
       { strmethodHead += ")"; }
      ({ String aux2; } aux2 = throwsClause { strmethodHead += " "+aux2; }
      )?
	;

throwsClause returns [ String strthrowsClause ] { strthrowsClause = ""; }
   :  #( "throws"
         { Vector ts = new Vector();
           String aux; }
         ( aux = identifier { ts.add(aux); })*
         { if(!ts.isEmpty())
           {
              strthrowsClause = "throws ";
              Iterator i = ts.iterator();
              while(i.hasNext())
              {
                 String t = (String) i.next();
                 strthrowsClause += t;
                 if(i.hasNext())
                    strthrowsClause += ",";
              }
           }
         }
         { if(!ts.isEmpty())
           {
              String exceptions = "";
              Iterator i = ts.iterator();
              while(i.hasNext())
              {
                 String t = (String) i.next();
                 exceptions += t;
                 if(i.hasNext())
                    exceptions += ",";
              }
              setOperationExceptions(exceptions);
           }
         }
       )
	;

identifier returns [ String strident ] { strident = ""; }
   :  tkident1:IDENT { strident = tkident1.getText(); }
   |  #( DOT { String aux; } aux=identifier tkident2:IDENT { strident = aux+"."+tkident2.getText(); })
	;

identifierStar returns [ String stridentifierStar ] { stridentifierStar = ""; }
   :  tkident1:IDENT { stridentifierStar = tkident1.getText(); }
   |  #( DOT { String aux1,aux2; }
         aux1=identifier
         (tkident2:STAR { aux2 = tkident2.getText(); }
         |tkident3:IDENT { aux2 = tkident3.getText(); }
         )
         { stridentifierStar = aux1+"."+aux2; }
       )
	;

ctorSList returns [ String strctorSList ] { strctorSList = ""; }
	:	#( SLIST
         ( { String aux1; } aux1 = ctorCall { strctorSList += getTab()+aux1+"\r\n"; } )?
         ( { String aux2; } aux2 = stat { strctorSList += getTab()+aux2+"\r\n"; } )*
       )
	;

slist returns [ String strslist ] { strslist = new String(); }
   :  #( SLIST
         ( { String strstat; }
           strstat = stat
           { strslist += getTab() + strstat + "\r\n"; }
         )*
       )
	;

stat returns [ String strstat ] { strstat = "undefined statement"; }
   : strstat = typeDefinition
   | strstat = variableDef { strstat += ";"; }
   | strstat = reservedVariableDef { strstat += ";"; }
   | strstat = expression { strstat += ";"; }
   |  #( LABELED_STAT tkident:IDENT { String aux; } aux=stat
         { strstat = tkident.getText()+" : "+aux; }
       )
   |  #( "if"
         { String aux1,aux2,aux3=null; }
         aux1=expression
         aux2=stat
         (aux3=stat)?
         { strstat = "if("+aux1+") " + aux2;
           if(aux3!=null)
              strstat += "\r\n"+getTab()+"else "+aux3;
         }
       )
   |  #( "for"
         { String aux1="",aux2="",aux3="",aux4; }
         #(FOR_INIT (aux1 = variableDef | aux1 = reservedVariableDef | aux1 = elist)?)
         #(FOR_CONDITION (aux2 = expression)?)
         #(FOR_ITERATOR (aux3 = elist)?)
         aux4 = stat
         { strstat = "for("+aux1+";"+aux2+";"+aux3+") "+aux4; }
       )
   |  #( "while"
         { String aux1,aux2; }
         aux1 = expression
         aux2 = stat
         { strstat = "while("+aux1+") "+aux2; }
       )
   |  #( "do"
         { String aux1,aux2; }
         aux1 = stat
         aux2 = expression
         { strstat = "do "+aux1+" while("+aux2+");"; }
       )
   |  #( "break"
         { strstat = "break"; }
         (tkbreak:IDENT { strstat += " "+tkbreak.getText(); })?
         { strstat += ";"; }
       )
   |  #( "continue"
         { strstat = "continue"; }
         (tkcontinue:IDENT { strstat += " "+tkcontinue.getText(); })?
         { strstat += ";"; }
       )
   |  #( "return"
         { strstat = "return"; }
         ({ String aux; } aux = expression { strstat += " "+aux; })?
         { strstat += ";"; }
       )
   |  #( "switch"
         { String aux; }
         aux = expression
         { strstat = "switch("+aux+")"; }
         { strstat += "\r\n"+getTab()+"{\r\n"; }
         { incTab(); }
         ({ String aux2; } aux2 = caseGroup { strstat+=aux2; })*
         { decTab(); }
         { strstat += getTab()+"}"; }
       )
   |  #( "throw"
         { String aux; }
         aux = expression
         { strstat = "throw "+aux+";"; }
       )
   |  #( "synchronized"
         { String aux1,aux2; }
         aux1 = expression
         aux2 = stat
         { strstat = "synchronized "+aux1+" "+aux2; }
       )
   |  strstat = tryBlock 
   |  { String aux; }
      { incTab(); }
      aux = slist
      { decTab(); }
      { strstat = "\r\n"+getTab()+"{\r\n"+aux+getTab()+"}"; }
	|	EMPTY_STAT
	;

caseGroup returns [ String strcaseGroup ] { strcaseGroup = ""; }
   :  #( CASE_GROUP
         (#( "case"
              { String aux; }
              aux = expression
              { strcaseGroup += getTab()+"case "+aux+" :\r\n"; }
            )
            | "default" { strcaseGroup = getTab()+"default :\r\n"; }
         )+
         { incTab(); }
         { String strslist; }
         { incTab(); }
         strslist = slist
         { decTab(); }
         { strcaseGroup += strslist; }
         { decTab(); }
       )
	;

tryBlock returns [ String strtryBlock ] { strtryBlock = ""; }
   :  #("try"
         { strtryBlock = "try\r\n"+getTab()+"{\r\n"; }
         { String strslist; }
         { incTab(); }
         strslist = slist
         { decTab(); }
         { strtryBlock += strslist; }
         { strtryBlock += getTab()+"}"; }
         ( { String aux; } aux = handler { strtryBlock += "\r\n"+aux; })*
         (#("finally"
            { String aux; }
            { incTab(); }
            aux = slist
            { decTab(); }
            { strtryBlock += "\r\n"+getTab()+"finally\r\n"; }
            { strtryBlock += getTab()+"{\r\n"; }
            { strtryBlock += aux; }
            { strtryBlock += getTab()+"}"; }
           )
         )?
       )
	;

handler returns [ String strtryBlock ] { strtryBlock = ""; }
   :  #("catch"
         { String aux1, aux2; }
         aux1 = parameterDef
         { incTab(); }
         aux2 = slist
         { decTab(); }
         { strtryBlock = getTab()+"catch("+aux1+")\r\n"; }
         { strtryBlock += getTab()+"{\r\n"; }
         { strtryBlock += aux2; }
         { strtryBlock += getTab()+"}"; }
       )
	;

elist returns [ String strelist ] { strelist = ""; }
   :  #( ELIST
         ({ String aux; } aux = expression { strelist += aux + ","; }
         { strelist = strelist.substring(0,strelist.length()-1); }
         )*
       )
	;

expression returns [ String strexpression ] { strexpression = ""; }
   :  #(EXPR strexpression = expr)
	;

expr returns [ String strexpr ] { strexpr = ""; String aux1,aux2,aux3; }
   :  #(QUESTION aux1=expr aux2=expr aux3=expr { strexpr = "("+aux1+" ? "+aux2+" : "+aux3+")"; }) // trinary operator
   |  #(ASSIGN aux1=expr aux2=expr { strexpr = aux1+" = "+aux2; })        // binary operators...
   |  #(PLUS_ASSIGN aux1=expr aux2=expr { strexpr = "("+aux1 + " += " + aux2+")"; })
   |  #(MINUS_ASSIGN aux1=expr aux2=expr { strexpr = "("+aux1 + " -= " + aux2+")"; })
   |  #(STAR_ASSIGN aux1=expr aux2=expr { strexpr = "("+aux1 + " *= " + aux2+")"; })
   |  #(DIV_ASSIGN aux1=expr aux2=expr { strexpr = "("+aux1 + " /= " + aux2+")"; })
   |  #(MOD_ASSIGN aux1=expr aux2=expr { strexpr = "("+aux1 + " %= " + aux2+")"; })
   |  #(SR_ASSIGN aux1=expr aux2=expr { strexpr = "("+aux1 + " >>= " + aux2+")"; })
   |  #(BSR_ASSIGN aux1=expr aux2=expr { strexpr = "("+aux1 + " >>>= " + aux2+")"; })
   |  #(SL_ASSIGN aux1=expr aux2=expr { strexpr = "("+aux1 + " <<= " + aux2+")"; })
   |  #(BAND_ASSIGN aux1=expr aux2=expr { strexpr = "("+aux1 + " &= " + aux2+")"; })
   |  #(BXOR_ASSIGN aux1=expr aux2=expr { strexpr = "("+aux1 + " ^= " + aux2+")"; })
   |  #(BOR_ASSIGN aux1=expr aux2=expr { strexpr = "("+aux1 + " |= " + aux2+")"; })
   |  #(LOR aux1=expr aux2=expr { strexpr = "("+aux1 + " || " + aux2+")"; })
   |  #(LAND aux1=expr aux2=expr { strexpr = "("+aux1 + " && " + aux2+")"; })
   |  #(BOR aux1=expr aux2=expr { strexpr = "("+aux1 + " | " + aux2+")"; })
   |  #(BXOR aux1=expr aux2=expr { strexpr = "("+aux1 + " ^ " + aux2+")"; })
   |  #(BAND aux1=expr aux2=expr { strexpr = "("+aux1 + " & " + aux2+")"; })
   |  #(NOT_EQUAL aux1=expr aux2=expr { strexpr = "("+aux1 + " != " + aux2+")"; })
   |  #(EQUAL aux1=expr aux2=expr { strexpr = "("+aux1 + " == " + aux2+")"; })
   |  #(LT aux1=expr aux2=expr { strexpr = "("+aux1 + " < " + aux2+")"; })
   |  #(GT aux1=expr aux2=expr { strexpr = "("+aux1 + " > " + aux2+")"; })
   |  #(LE aux1=expr aux2=expr { strexpr = "("+aux1 + " <= " + aux2+")"; })
   |  #(GE aux1=expr aux2=expr { strexpr = "("+aux1 + " >= " + aux2+")"; })
   |  #(SL aux1=expr aux2=expr { strexpr = "("+aux1 + " << " + aux2+")"; })
   |  #(SR aux1=expr aux2=expr { strexpr = "("+aux1 + " >> " + aux2+")"; })
   |  #(BSR aux1=expr aux2=expr { strexpr = "("+aux1 + " >>> " + aux2+")"; })
   |  #(PLUS aux1=expr aux2=expr { strexpr = "("+aux1 + " + " + aux2+")"; })
   |  #(MINUS aux1=expr aux2=expr { strexpr = "("+aux1 + " - " + aux2+")"; })
   |  #(DIV aux1=expr aux2=expr { strexpr = "("+aux1 + " / " + aux2+")"; })
   |  #(MOD aux1=expr aux2=expr { strexpr = "("+aux1 + " % " + aux2+")"; })
   |  #(STAR aux1=expr aux2=expr { strexpr = "("+aux1 + " * " + aux2+")"; })
   |  #(INC aux1=expr { strexpr = "("+"++" + aux1+")"; })
   |  #(DEC aux1=expr { strexpr = "("+"--" + aux1+")"; })
   |  #(POST_INC aux1=expr { strexpr = "("+aux1 + "++"+")"; })
   |  #(POST_DEC aux1=expr { strexpr = "("+aux1 + "--"+")"; })
   |  #(BNOT aux1=expr { strexpr = "("+"~" + aux1+")"; })
   |  #(LNOT aux1=expr { strexpr = "("+"!" + aux1+")"; })
   |  #("instanceof" aux1=expr aux2=expr { strexpr = "("+aux1 + " instanceof " + aux2+")"; })
   |  #(UNARY_MINUS aux1=expr { strexpr = "("+"-" + aux1+")"; })
   |  #(UNARY_PLUS aux1=expr { strexpr = "("+"+" + aux1+")"; })
   |  strexpr = primaryExpression 
	;

primaryExpression returns [ String strprimaryExpression ] { strprimaryExpression = ""; }
    :   tkident1:IDENT { strprimaryExpression = tkident1.getText(); }
    | #(  DOT
         ( { String aux1; }
            aux1=expr
            (  tkident2:IDENT { strprimaryExpression = aux1+"."+tkident2.getText(); }
            |  { String aux2; } aux2=arrayIndex { strprimaryExpression = aux1+"."+aux2; }
            |  "this" { strprimaryExpression = aux1+".this"; }
            |  "class" { strprimaryExpression = aux1+".class"; }
            |  #( "new" tkident3:IDENT { String aux3; } aux3=elist { strprimaryExpression = aux1+".new("+aux3+")"; } )
            |   "super" { strprimaryExpression = aux1+".super"; }
				)
         |  #(ARRAY_DECLARATOR { String aux4; } aux4=typeSpecArray { strprimaryExpression = aux4; })
         |  { String aux5; } aux5 = builtInType ("class" { aux5+=".class"; } )?{ strprimaryExpression = aux5; }
			)
		)
    |  strprimaryExpression = arrayIndex
    |  #(METHOD_CALL { String aux1,aux2; } aux1=primaryExpression aux2=elist { strprimaryExpression = aux1+"("+aux2+")"; })
    |  #(TYPECAST { String aux1,aux2; } aux1=typeSpec aux2=expr { strprimaryExpression = "("+aux1+")"+aux2; })
    |  strprimaryExpression = newExpression
    |  strprimaryExpression = constant
    |  "super" { strprimaryExpression = "super"; }
    |  "true" { strprimaryExpression = "true"; }
    |  "false" { strprimaryExpression = "false"; }
    |  "this" { strprimaryExpression = "this"; }
    |  "null" { strprimaryExpression = "null"; }
    |  strprimaryExpression = typeSpec
       // type name used with instanceof
    ;

ctorCall returns [ String strctorCall ] { strctorCall = ""; }
   :  #( CTOR_CALL { String aux1; } aux1=elist { strctorCall += "this("+aux1+");"; } )
	|	#( SUPER_CTOR_CALL
         ( { String aux2; } aux2=elist { strctorCall += "super("+aux2+");"; }
         | { String aux3,aux4; } aux3=primaryExpression aux4=elist { strctorCall += aux3+".super("+aux4+");"; }
			)
		 )
	;

arrayIndex returns [ String strarrayIndex ] { strarrayIndex = ""; }
   :  #(INDEX_OP { String aux1,aux2; } aux1=primaryExpression aux2=expression { strarrayIndex = aux1+"["+aux2+"]"; })
	;

constant returns [ String strconstant ] { strconstant = ""; }
    :   tk1:NUM_INT { strconstant = tk1.getText(); }
    |   tk2:CHAR_LITERAL { strconstant = tk2.getText(); }
    |   tk3:STRING_LITERAL { strconstant = tk3.getText(); }
    |   tk4:NUM_FLOAT { strconstant = tk4.getText(); }
    |   tk5:NUM_DOUBLE { strconstant = tk5.getText(); }
    |   tk6:NUM_LONG { strconstant = tk6.getText(); }
    ;

newExpression returns [ String strnewExpression ] { strnewExpression = ""; }
   :  #( "new" { String aux1; } aux1=type { strnewExpression = "new "+aux1; }
         (  { String aux2,aux3; } aux2=newArrayDeclarator { strnewExpression += aux2; }(aux3=arrayInitializer { strnewExpression += " = "+aux3; })?
         |  { String aux4,aux5; } aux4=elist { strnewExpression += "("+aux4+")"; } (aux5=objBlock { strnewExpression += aux5; })?
			)
		)
			
	;

newArrayDeclarator returns [ String strnewArrayDeclarator ] { strnewArrayDeclarator = ""; }
   :  #( ARRAY_DECLARATOR  { String aux1=null,aux2=null; } ( aux1=newArrayDeclarator)? (aux2=expression)? { strnewArrayDeclarator = aux1+"["+aux2+"]"; })
	;
